/***                                                                      ***\

    logresolve 3.0


    Robert R. Collins / rcollins@x86.org / http://www.x86.org, May  1, 1998


    Usage: logresolve [arguments] < access_log > new_log        (v1.0 compliant)
    - or - logresolve [arguments] [< access_log] [> new_log]    (v2.0 compliant)

    Default Behavior:   Identical to logresolve version 1.0.
                        Input is taken from STDIN.
                        Output is directed to STDOUT.
                        Statistics are directed to STDERR.

    Arguments (if specified):  specifying any other arguments causes
                               statistics to be printed to STDERR (or ERRFILE).
        -i INFILE  :    Indicates an input file to be read.  When this argument
                        is given, input is read from the input file, not STDIN.
                        This argument overrides the default behavior.

        -o OUTFILE :    Indicates an output file to be written.  When this argument
                        is given, output is directed to the output file, not STDOUT.
                        This argument overrides the default behavior.


    Notes:

    To generate meaningful statistics from an HTTPD log file, it's good
    to have the domain name of each machine that accessed your site, but
    doing this on the fly can slow HTTPD down.

    Compiling NCSA HTTPD with the -DMINIMAL_DNS flag turns IP#->hostname
    resolution off. Before running your stats program, just run your log
    file through this program (logresolve) and all of your IP numbers will
    be resolved into hostnames (where possible).

    logresolve takes an HTTPD access log (in the COMMON log file format,
    or any other format that has the IP number/domain name as the first
    field for that matter), and outputs the same file with all of the
    domain names looked up. Where no domain name can be found, the IP
    number is left in.

    To minimize impact on your nameserver, logresolve has its very own
    internal direct-mapped cache. This means that each IP number will only
    be looked up the first time it is found in the log file. As noted
    above, giving any command-line arguments to logresolve (anything at
    all!) will give you some statistics on the log file printed to STDERR.

\***                                                                      ***/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "log.h"
#include "logdata.h"

int
main(argc, argv)
    int             argc;
    char           *argv[];
{
    char            infile[MAXLINE]     = {""};
    char        outfile[MAXLINE]        = {""};
    char        errfile[MAXLINE]        = {""};
    char        cachefile[MAXLINE]      = {""};
    char            line[MAXLINE]       = {""};
    unsigned long   i;
    FILE           *in, *out;
    int             response;

    in  = stdin;
    out = stdout;

    for(i = argc, argc = 1; argc < i; argc ++)
    {
        switch(argv[argc][0])
        {
            case '-' :
                switch(argv[argc][1])
                {
                    case 'i' :          /* Capture INPUT file           */
                        strcpy(infile, argv[++argc]);
                        Status.UseInFile    = 1;
                        Status.PrintStats   = 1;
                        continue;

                    case 'o' :          /* Capture OUTPUT file           */
                        strcpy(outfile, argv[++argc]);
                        Status.UseOutFile = 1;
                        Status.PrintStats   = 1;
                        continue;

                    case '-' :                          /* Print help message.                  */
                    case 'h' :
                    case '?' :
                        printf("\n");
                        printf("logresolve -- Resolve IP address in a log file to domain names.\n\n");
                        printf("Usage:  logresolve -i infile -o outfile [-ci cachefile_in] [-co cachefile_out]\n");
                        printf("   -i infile        : Specify input file .\n");
                        printf("   -o outfile       : Specify output file.\n");
                        printf("\n");
                        fflush(stdout);
                        return (0);
                }
                    
            default :          /* Capture INPUT file           */
                printf("Invalid command-line option: %s.\n", argv[argc]);
                exit(-1);
        }
    }

    /* If an input cachefile has been specified, but not an output      *
     * cachefile, then the output cache file should assume the same     *
     * name as the input cachefile.                                     */
    if(Status.UseCacheFile && !Status.UseErrFile)
    {
        strcpy(errfile, cachefile);
        Status.UseErrFile = 1;
    }

    /* Open the input and output files.  We can't open the cache files  *
     * here, because we might want to use the same filename for the     *
     * input and output DNS cache files.                                */
    if(Status.UseInFile)
    {
        if((in = fopen(infile,"r")) == NULL) 
        {
            fprintf(stderr,"Input file not found:  %s\n",infile);
            exit(1);
        }
    }

    if(Status.UseOutFile)
    {
        if((out = fopen(outfile,"w")) == NULL)
        {
            fprintf(stderr,"Can't open output file:  %s\n",outfile);
            exit(1);
        }
    }

    /* Process the input log file.  The input log file will be handled  *
     * in a psuedo nested loop.  MAXFILLSIZE lines are always read,     *
     * except when we reach the end-of-file.  During normal processing, *
     * the outer loop counter will always equal MAXFILLSIZE.  When EOF  *
     * is reached, our  outer loop counter indicates how many lines     *
     * remain in the input log file (modulo MAXFILLSIZE).  This simple  *
     * optimization makes end-of-file processing very easy, and without *
     * creating a special case to handle it.                            */
    fprintf(stderr, "Counting input file entries...\n");
    fflush(stderr);
    while(feof(in) == 0)
    {
        if(getline(in, line, MAXLINE))
            break;
        statistics.entries ++;
    }

    ip_array = malloc(statistics.entries * sizeof(struct ip_array_s));
    if(ip_array == NULL)
    {
        fprintf(stderr, "Unable to malloc memory for IP Array.\n");
        fflush(stderr);
        exit(-1);
    }

    fprintf(stderr, "Extracting IP numbers...\n");
    fflush(stderr);
    fseek(in, 0L, SEEK_SET);
    for(i = 0; i < statistics.entries; i ++)
    {
        if(getline(in, line, MAXLINE))
            break;
        ip_array[i].sortindex = i;
        ExtractIPNUM(line, &ip_array[i]);

    }

    fprintf(stderr, "Sorting by IP numbers...\n");
    fflush(stderr);
    SortEntriesByIPNUM(statistics.entries);

    fprintf(stderr, "Counting unique IP addresses...\n");
    fflush(stderr);
    CountUniqueIPNUMs(statistics.entries);

    dnscache = malloc(statistics.ipnums * sizeof(struct dnsrec));
    if(dnscache == NULL)
    {
        fprintf(stderr, "Unable to malloc memory for DNS array...\n");
        fflush(stderr);
        exit(-1);
    }

    fprintf(stderr, "Resolving cached entries for redundant DNS lookups...\n");
    fflush(stderr);
    FindCachedEntries(statistics.entries);

    fprintf(stderr, "Resolving IP addresses with DNS lookups...\n");
    fflush(stderr);
    for(i = 0; i < statistics.ipnums; i++)
    {
        response = DNSLookup(&dnscache[i]);
        (response ? statistics.resolved ++ : statistics.noname ++);
    }

    fprintf(stderr, "Writing to output file...\n");
    fflush(stderr);
    fseek(in, 0L, SEEK_SET);
    for(i = 0; i < statistics.entries; i ++)
    {
        if(getline(in, line, MAXLINE))
            break;
        if(ip_array[i].iplong)
        {
            fprintf(out, "%s %s\n", ip_array[i].dnsname, ip_array[i].lineout);
        }
        else
        {
            fprintf(out, "%s\n", line);
        }
    }

    free(ip_array);
    free(dnscache);

    if(Status.PrintStats)
    {
        PrintStats(stderr);
    }


    if(Status.UseInFile)
        fclose(in);

    if(Status.UseOutFile)
        fclose(out);

    return (0);
}
